"""
This file defines the class definition of a Karel world.

The sub header comment defines important notes about the Karel
world file format.

General Notes About World Construction
- Streets run EAST-WEST (rows)
- Avenues run NORTH-SOUTH (columns)

World File Constraints:
- World file should specify one component per line in the format
  KEYWORD: PARAMETERS
- Any lines with no colon delimiter will be ignored
- The accepted KEYWORD, PARAMETER combinations are as follows:
    - Dimension: (num_avenues, num_streets)
    - Wall: (avenue, street); direction
    - Beeper: (avenue, street) count
    - Karel: (avenue, street); direction
    - Color: (avenue, street); color
    - Speed: delay
    - BeeperBag: num_beepers
- Multiple parameter values for the same keyword should be separated by a semicolon
- All numerical values (except delay) must be expressed as ints. The exception
  to this is that the number of beepers can also be INFINITY
- Any specified color values must be valid TKinter color strings, and are limited
  to the set of colors
- Direction is case-insensitive and can be one of the following values:
    - East
    - West
    - North
    - South

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
"""
from __future__ import annotations

import collections
import copy
import re
import sys
from enum import Enum, unique
from pathlib import Path
from typing import Any, NamedTuple

INFINITY = -1
COLOR_MAP = {
    "Red": "red",
    "Black": "black",
    "Cyan": "cyan",
    "Dark Gray": "gray30",
    "Gray": "gray55",
    "Green": "green",
    "Light Gray": "gray80",
    "Magenta": "magenta3",
    "Orange": "orange",
    "Pink": "pink",
    "White": "snow",
    "Blue": "blue",
    "Yellow": "yellow",
}
INIT_SPEED = 50
VALID_WORLD_KEYWORDS = [
    "dimension",
    "wall",
    "beeper",
    "karel",
    "speed",
    "beeperbag",
    "color",
]
KEYWORD_DELIM = ":"
PARAM_DELIM = ";"
DEFAULT_WORLD_FILE = "default_world.w"


class KarelWorld:
    def __init__(self, world_file: str) -> None:
        """
        Karel World constructor
        Parameters:
            world_file: filename containing the initial state of Karel's world
        """
        self.world_file = self.process_world(world_file)

        # Map of beeper locations to the count of beepers at that location
        self.beepers: dict[tuple[int, int], int] = collections.defaultdict(int)

        # Map of corner colors, defaults to ""
        self.corner_colors: dict[tuple[int, int], str] = collections.defaultdict(
            lambda: ""
        )

        # Set of Wall objects placed in the world
        self.walls: set[Wall] = set()

        # Dimensions of the world
        self.num_streets = 1
        self.num_avenues = 1

        # Initial Karel state saved to enable world reset
        self.karel_start_location = (1, 1)
        self.karel_start_direction = Direction.EAST
        self.karel_start_beeper_count = 0

        # Initial speed slider setting
        self.init_speed = INIT_SPEED

        # If a world file has been specified, load world details from the file
        if self.world_file:
            self.load_from_file()

        # Save initial beeper state to enable world reset
        self.init_beepers = copy.deepcopy(self.beepers)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, KarelWorld):
            return (
                self.beepers == other.beepers
                and self.walls == other.walls
                and self.num_streets == other.num_streets
                and self.num_avenues == other.num_avenues
                and self.corner_colors == other.corner_colors
            )
        return NotImplemented

    @staticmethod
    def process_world(world_file: str) -> Path:
        """
        If no world_file is provided, use default world.
        Find world file that matches program name in the current worlds/ directory.
        If not found, search the provided default worlds directory.
        """
        default_worlds_path = Path(__file__).absolute().parent / "worlds"
        if not world_file:
            default_world = default_worlds_path / DEFAULT_WORLD_FILE
            if default_world.is_file():
                print("Using default world...")
                return default_world
            raise FileNotFoundError(
                f"Default world cannot be found in: {default_worlds_path}\n"
                "Please raise an issue on the stanfordkarel GitHub."
            )

        world_filepath = Path(world_file)
        if world_filepath.is_file():
            return world_filepath

        worlds_folder = Path("worlds")
        for worlds_path in (worlds_folder, default_worlds_path):
            if worlds_path.is_dir():
                full_world_path = worlds_path / f"{world_file}.w"
                if full_world_path.is_file():
                    return full_world_path

        if not worlds_folder.is_dir():
            print("Could not find worlds/ folder in current directory.\n")

        sys.tracebacklimit = 0
        available_worlds = "\n".join(
            [f"  {world.stem}" for world in default_worlds_path.glob("*.w")]
        )
        raise FileNotFoundError(
            "The specified file was not one of the provided worlds.\n"
            "Please store custom worlds in a folder named worlds/, "
            f"or use a world listed below:\n{available_worlds}"
            "\nPass the default world as a parameter in run_karel_program().\n"
            "    e.g. run_karel_program('checkerboard_karel')"
        )

    @staticmethod
    def get_alt_wall(wall: Wall) -> Wall:
        if wall.direction == Direction.NORTH:
            return Wall(wall.avenue, wall.street + 1, Direction.SOUTH)
        if wall.direction == Direction.SOUTH:
            return Wall(wall.avenue, wall.street - 1, Direction.NORTH)
        if wall.direction == Direction.EAST:
            return Wall(wall.avenue + 1, wall.street, Direction.WEST)
        if wall.direction == Direction.WEST:
            return Wall(wall.avenue - 1, wall.street, Direction.EAST)
        raise ValueError

    @staticmethod
    def parse_parameters(keyword: str, param_str: str) -> dict[str, Any]:
        params: dict[str, Any] = {}
        for param in param_str.split(PARAM_DELIM):
            param = param.strip()

            # check to see if parameter encodes a location
            coordinate = re.match(r"\((\d+),\s*(\d+)\)", param)
            if coordinate:
                # avenue, street
                params["location"] = int(coordinate.group(1)), int(coordinate.group(2))
                continue

            # check to see if the parameter is a direction value
            if param in (d.value for d in Direction):
                params["direction"] = Direction(param)

            # check to see if parameter encodes a numerical value or color string
            elif keyword == "color":
                if param.title() not in COLOR_MAP:
                    raise ValueError(
                        f"Error: {param} is invalid parameter for {keyword}."
                    )
                params["color"] = param.title()

            # handle the edge case where Karel has infinite beepers
            elif param in ("infinity", "infinite") and keyword == "beeperbag":
                params["val"] = INFINITY

            # float values are only valid for the speed parameter.
            elif keyword == "speed":
                try:
                    params["val"] = int(100 * float(param))
                except ValueError as e:
                    raise ValueError(
                        f"Error: {param} is an invalid parameter for {keyword}."
                    ) from e

            # must be a digit then
            elif param.isdigit():
                params["val"] = int(param)

            else:
                raise ValueError(f"Error: {param} is invalid parameter for {keyword}.")
        return params

    def load_from_file(self) -> None:
        with open(self.world_file, encoding="utf-8") as f:
            for i, line in enumerate(f):
                # Ignore blank lines and lines with no comma delineator
                line = line.strip()
                if not line:
                    continue

                if KEYWORD_DELIM not in line:
                    print(f"Incorrectly formatted - ignoring line {i} of file: {line}")
                    continue

                keyword, param_str = line.lower().split(KEYWORD_DELIM)

                # only accept valid keywords as defined in world file spec
                # TODO: add error detection for keywords with insufficient parameters
                params = self.parse_parameters(keyword, param_str)

                # handle all different possible keyword cases
                if keyword == "dimension":
                    # set world dimensions based on location values
                    self.num_avenues, self.num_streets = params["location"]

                elif keyword == "wall":
                    # build a wall at the specified location
                    (avenue, street), direction = (
                        params["location"],
                        params["direction"],
                    )
                    self.walls.add(Wall(avenue, street, direction))

                elif keyword == "beeper":
                    # add the specified number of beepers to the world
                    self.beepers[params["location"]] += params["val"]

                elif keyword == "karel":
                    # Give Karel initial state values
                    self.karel_start_location = params["location"]
                    self.karel_start_direction = params["direction"]

                elif keyword == "beeperbag":
                    # Set Karel's initial beeper bag count
                    self.karel_start_beeper_count = params["val"]

                elif keyword == "speed":
                    # Set delay speed of program execution
                    self.init_speed = params["val"]

                elif keyword == "color":
                    # Set corner color to be specified color
                    self.corner_colors[params["location"]] = params["color"]

                else:
                    print(f"Invalid keyword - ignoring line {i} of world file: {line}")

    def set_karel_start_location(self, avenue: int, street: int) -> None:
        self.karel_start_location = (avenue, street)

    def set_karel_start_direction(self, direction: Direction) -> None:
        self.karel_start_direction = direction

    def set_karel_start_beeper_count(self, beeper_count: int) -> None:
        self.karel_start_beeper_count = beeper_count

    def add_beeper(self, avenue: int, street: int) -> None:
        self.beepers[(avenue, street)] += 1

    def remove_beeper(self, avenue: int, street: int) -> None:
        if self.beepers[(avenue, street)] > 0:
            self.beepers[(avenue, street)] -= 1

    def add_wall(self, wall: Wall) -> None:
        alt_wall = self.get_alt_wall(wall)
        if wall not in self.walls and alt_wall not in self.walls:
            self.walls.add(wall)

    def remove_wall(self, wall: Wall) -> None:
        alt_wall = self.get_alt_wall(wall)
        if wall in self.walls:
            self.walls.remove(wall)
        if alt_wall in self.walls:
            self.walls.remove(alt_wall)

    def paint_corner(self, avenue: int, street: int, color: str) -> None:
        self.corner_colors[(avenue, street)] = color

    def corner_color(self, avenue: int, street: int) -> str:
        return self.corner_colors[(avenue, street)]

    def reset_corner(self, avenue: int, street: int) -> None:
        self.beepers[(avenue, street)] = 0
        self.corner_colors[(avenue, street)] = ""

    def wall_exists(self, avenue: int, street: int, direction: Direction) -> bool:
        wall = Wall(avenue, street, direction)
        return wall in self.walls

    def in_bounds(self, avenue: int, street: int) -> bool:
        return 0 < avenue <= self.num_avenues and 0 < street <= self.num_streets

    def reset_world(self) -> None:
        """Reset initial state of beepers in the world"""
        self.beepers = copy.deepcopy(self.init_beepers)
        self.corner_colors = collections.defaultdict(lambda: "")

    def reload_world(self, filename: str | None = None) -> None:
        """Reloads world using constructor."""
        # TODO To fix this, we need to figure out why the constructor does not reset
        # everything. Use New World to test.
        # pylint: disable=unnecessary-dunder-call
        self.__init__(filename)  # type: ignore[misc]

    def save_to_file(self, filename: Path) -> None:
        # First, output dimensions of world
        output = f"Dimension: ({self.num_avenues}, {self.num_streets})\n"

        # Next, output all walls
        for wall in sorted(self.walls):
            output += f"Wall: ({wall.avenue}, {wall.street}); {wall.direction.value}\n"

        # Next, output all beepers
        for (x, y), count in sorted(self.beepers.items()):
            output += f"Beeper: ({x}, {y}); {count}\n"

        # Next, output all color information
        for (x, y), color in sorted(self.corner_colors.items()):
            if color:
                output += f"Color: ({x}, {y}); {color}\n"

        # Next, output Karel information
        output += (
            f"Karel: {self.karel_start_location}; {self.karel_start_direction.value}\n"
        )

        # Finally, output beeperbag info
        beeper_output = (
            self.karel_start_beeper_count
            if self.karel_start_beeper_count >= 0
            else "INFINITY"
        )
        output += f"BeeperBag: {beeper_output}\n"

        with open(filename, "w", encoding="utf-8") as f:
            f.write(output)


@unique
class Direction(Enum):
    EAST = "east"
    SOUTH = "south"
    WEST = "west"
    NORTH = "north"

    def __lt__(self, other: object) -> bool:
        """Required to sort Directions."""
        if isinstance(other, Direction):
            return self.value < other.value
        return NotImplemented

    def __repr__(self) -> str:
        return str(self.value)


class Wall(NamedTuple):
    """Note that the World Editor only uses West & South to denote wall directions."""

    avenue: int
    street: int
    direction: Direction



"""
This file defines the necessary functions and definitions that students must
import in order to be able to write a new Karel program. Any new Karel file
must include the following line:

from stanfordkarel import *

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
"""
import sys
import tkinter as tk
from pathlib import Path

from .karel_application import KarelApplication
from .karel_program import KarelProgram

# The following function definitions are defined as stubs so that IDEs can recognize
# the function definitions in student code. These names are re-bound upon program
# execution to asscoiate their behavior to the one particular Karel object located
# in a given world.


def move() -> None:
    pass


def turn_left() -> None:
    pass


def put_beeper() -> None:
    pass


def pick_beeper() -> None:
    pass


def front_is_clear() -> bool:
    pass


def front_is_blocked() -> bool:
    pass


def left_is_clear() -> bool:
    pass


def left_is_blocked() -> bool:
    pass


def right_is_clear() -> bool:
    pass


def right_is_blocked() -> bool:
    pass


def beepers_present() -> bool:
    pass


def no_beepers_present() -> bool:
    pass


def beepers_in_bag() -> bool:
    pass


def no_beepers_in_bag() -> bool:
    pass


def facing_north() -> bool:
    pass


def not_facing_north() -> bool:
    pass


def facing_east() -> bool:
    pass


def not_facing_east() -> bool:
    pass


def facing_west() -> bool:
    pass


def not_facing_west() -> bool:
    pass


def facing_south() -> bool:
    pass


def not_facing_south() -> bool:
    pass


def paint_corner(color: str) -> None:
    del color


def corner_color_is(color: str) -> bool:
    del color
    return True


# Defined constants for ease of use by students when coloring corners
RED = "Red"
BLACK = "Black"
CYAN = "Cyan"
DARK_GRAY = "Dark Gray"
GRAY = "Gray"
GREEN = "Green"
LIGHT_GRAY = "Light Gray"
MAGENTA = "Magenta"
ORANGE = "Orange"
PINK = "Pink"
WHITE = "White"
BLUE = "Blue"
YELLOW = "Yellow"
BLANK = ""


def run_karel_program(world_file: str = "") -> None:
    # Extract the name of the file the student is executing
    student_code_file = Path(sys.argv[0])

    # Special case - if filename matches a specified world name,
    # Set the default world to the world with that name.
    # I personally recommend removing this functionality completely.
    if (
        world_file == ""
        and (
            Path(__file__).absolute().parent
            / "worlds"
            / student_code_file.with_suffix(".w").name
        ).is_file()
    ):
        world_file = student_code_file.stem

    # Create Karel and assign it to live in the newly created world
    karel = KarelProgram(world_file)

    # Initialize root Tk Window and spawn Karel application
    root = tk.Tk()
    app = KarelApplication(karel, student_code_file, master=root)
    app.mainloop()



"""
This file defines the logic to add suggestions to exceptions.

Original Author: Tyler Yep
Credits: Sylvain Desodt
License: MIT
Version: 1.0.0
Email: tyep@cs.stanford.edu
"""
from __future__ import annotations

import difflib
import itertools
import re
from types import FrameType, TracebackType
from typing import Any, Callable, Iterator

# To be used in `get_suggestions_for_exception`.
SUGGESTION_FUNCTIONS: dict[Any, list[Any]] = {}
VAR_NAME = r"[^\d\W]\w*"
NAMENOTDEFINED_RE = rf"^(?:global )?name '(?P<name>{VAR_NAME})' is not defined$"


def merge_dict(*dicts: dict[str, Any]) -> dict[str, list[str]]:
    """
    Merge dicts and return a dictionary mapping key to list of values.
    Order of the values corresponds to the order of the original dicts.
    """
    ret: dict[str, list[str]] = {}
    for dict_ in dicts:
        for key, val in dict_.items():
            ret.setdefault(key, []).append(val)
    return ret


def add_scope_to_dict(dict_: dict[str, Any], scope: str) -> dict[str, tuple[Any, str]]:
    """Convert name:obj dict to name: (obj, scope) dict."""
    return {k: (v, scope) for k, v in dict_.items()}


def get_objects_in_frame(frame: FrameType) -> dict[str, list[str]]:
    """Get objects defined in a given frame.

    This includes variable, types, builtins, etc.
    The function returns a dictionary mapping names to a (non empty)
    list of ScopedObj objects in the order following the LEGB Rule.
    """
    return merge_dict(
        add_scope_to_dict(frame.f_locals, "local"),
        add_scope_to_dict(frame.f_globals, "global"),
        add_scope_to_dict(frame.f_builtins, "builtin"),
    )


def register_suggestion_for(
    error_type: type[Exception], regex: str
) -> Callable[..., Any]:
    """Decorator to register a function to be called to get suggestions.

    Parameters correspond to the fact that the registration is done for a
    specific error type and if the error message matches a given regex
    (if the regex is None, the error message is assumed to match before being
    retrieved).

    The decorated function is expected to yield any number (0 included) of
    suggestions (as string).
    The parameters are: (value, frame, groups):
     - value: Exception object
     - frame: Last frame of the traceback (may be None when the traceback is
        None which happens only in edge cases)
     - groups: Groups from the error message matched by the error message.
    """

    def internal_decorator(func: Callable[..., list[str]]) -> Callable[..., Any]:
        def registered_function(value: Exception, frame: FrameType) -> list[str]:
            if regex is None:
                return func(value, frame, [])
            error_msg = value.args[0]
            match = re.match(regex, error_msg)
            if match:
                return func(value, frame, match.groups())
            return []

        SUGGESTION_FUNCTIONS.setdefault(error_type, []).append(registered_function)
        return func  # return original function

    return internal_decorator


@register_suggestion_for(NameError, NAMENOTDEFINED_RE)
def suggest_name_not_defined(
    value: Exception, frame: FrameType, groups: tuple[str]
) -> itertools.chain[str]:
    """Get the suggestions for name in case of NameError."""
    del value
    (name,) = groups
    objs = get_objects_in_frame(frame)
    return itertools.chain(suggest_name_as_name_typo(name, objs))


def get_suggestions_for_exception(
    value: Exception, traceback: TracebackType | None
) -> str:
    """Get suggestions for an exception."""
    frame = get_last_frame(traceback)
    suggestions = itertools.chain.from_iterable(
        func(value, frame)
        for error_type, functions in SUGGESTION_FUNCTIONS.items()
        if isinstance(value, error_type)
        for func in functions
    )
    if suggestions:
        return f". Did you mean {', '.join(list(suggestions))}?"
    return ""


def get_close_matches(word: str, possibilities: list[str]) -> list[str]:
    """
    Return a list of the best "good enough" matches.

    Wrapper around difflib.get_close_matches() to be able to
    change default values or implementation details easily.
    """
    return [
        w for w in difflib.get_close_matches(word, possibilities, 3, 0.7) if w != word
    ]


def suggest_name_as_name_typo(name: str, objdict: dict[str, Any]) -> Iterator[str]:
    """Suggest that name could be a typo (misspelled existing name).

    Example: 'foobaf' -> 'foobar'.
    """
    for word in get_close_matches(name, list(objdict)):
        yield f"'{word}'"  # + " (" + objdict[name][0].scope + ")"


def add_string_to_exception(value: Exception, string: str) -> None:
    """
    Add string to the exception parameter.

    The point is to have the string visible when the exception is printed or converted
    to string - may it be via `str()`, `repr()` or when the exception is uncaught and
    displayed (which seems to use `str()`). In an ideal world, one just needs to update
    `args` but apparently it is not enough for SyntaxError, IOError, etc where other
    attributes (`msg`, `strerror`, `reason`, etc) are to be updated too (for `str()`,
    not for `repr()`). Also, elements in args might not be strings or args might me
    empty so we add to the first string and add the element otherwise.
    """
    if not isinstance(value.args, tuple):
        raise RuntimeError
    if string:
        lst_args = list(value.args)
        for i, arg in enumerate(lst_args):
            if isinstance(arg, str):
                lst_args[i] = arg + string
                break
        else:
            # if no string arg, add the string anyway
            lst_args.append(string)
        value.args = tuple(lst_args)
        for attr in ("msg", "strerror", "reason"):
            attrval = getattr(value, attr, None)
            if attrval is not None:
                setattr(value, attr, attrval + string)


def get_last_frame(traceback: TracebackType | None) -> FrameType | None:
    """Extract last frame from a traceback."""
    # In some rare case, the given traceback might be None
    if traceback is None:
        return None
    while traceback.tb_next:
        traceback = traceback.tb_next
    return traceback.tb_frame


def add_did_you_mean(e: Exception) -> None:
    """Hook to be substituted to sys.excepthook to enhance exceptions."""
    if isinstance(e, NameError):
        suggestions = get_suggestions_for_exception(e, e.__traceback__)
        add_string_to_exception(e, suggestions)



"""
This file defines the GUI for running Karel programs.

Original Author: Nicholas Bowman
Credits: Kylie Jue, Tyler Yep
License: MIT
Version: 1.0.0
Email: nbowman@stanford.edu
Date of Creation: 10/1/2019
"""
from __future__ import annotations

import importlib.util
import inspect
import tkinter as tk
import traceback as tb
from pathlib import Path
from time import sleep
from tkinter.filedialog import askopenfilename
from tkinter.messagebox import showwarning
from types import FrameType, ModuleType
from typing import Any, Callable, cast

from .didyoumean import add_did_you_mean
from .karel_canvas import DEFAULT_ICON, LIGHT_GREY, PAD_X, PAD_Y, KarelCanvas
from .karel_program import KarelException, KarelProgram


class StudentModule(ModuleType):
    move: Any
    turn_left: Any
    put_beeper: Any
    pick_beeper: Any
    paint_corner: Any

    @staticmethod
    def main() -> None:
        raise NotImplementedError


class StudentCode:
    """
    This process extracts a module from an arbitary file that contains student code.
    https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    """

    def __init__(self, code_file: Path) -> None:
        if not code_file.is_file():
            raise FileNotFoundError(f"{code_file} could not be found.")

        self.module_name = code_file.stem
        spec = importlib.util.spec_from_file_location(
            self.module_name, code_file.resolve()
        )
        assert spec is not None
        try:
            module_loader = spec.loader
            assert module_loader is not None
            mod = cast(StudentModule, importlib.util.module_from_spec(spec))
            self.mods: list[StudentModule] = [mod]
            module_loader.exec_module(mod)
            # Go through attributes to find imported modules
            for name in dir(mod):
                module = cast(StudentModule, getattr(mod, name))
                if isinstance(module, ModuleType):
                    assert module.__file__ is not None
                    code_file_path = Path(module.__file__)
                    # Only execute modules outside of this directory
                    if code_file_path.parent != Path(__file__).resolve().parent:
                        self.mods.append(module)
                        spec = importlib.util.spec_from_file_location(
                            name, code_file_path.resolve()
                        )
                        module_loader.exec_module(module)
        except SyntaxError as e:
            # Since we don't start the GUI until after we parse the student's code,
            # SyntaxErrors behave normally. However, if the syntax error is somehow
            # not caught at parse time, we should forward the error message to console.
            print(e)
            raise

        # Do not proceed if the student has not defined a main function.
        if not hasattr(self.mods[0], "main"):
            raise RuntimeError(
                "Couldn't find the main() function. Are you sure you have one?"
            )

    def __repr__(self) -> str:
        return "\n".join([inspect.getsource(mod) for mod in self.mods])

    def inject_namespace(self, karel: KarelProgram) -> None:
        """
        This function is responsible for doing some Python hackery
        that associates the generic commands the student wrote in their
        file with specific commands relating to the Karel object that exists
        in the world.
        """
        functions_to_override = [
            "move",
            "turn_left",
            "pick_beeper",
            "put_beeper",
            "facing_north",
            "facing_south",
            "facing_east",
            "facing_west",
            "not_facing_north",
            "not_facing_south",
            "not_facing_east",
            "not_facing_west",
            "front_is_clear",
            "beepers_present",
            "no_beepers_present",
            "beepers_in_bag",
            "no_beepers_in_bag",
            "front_is_blocked",
            "left_is_blocked",
            "left_is_clear",
            "right_is_blocked",
            "right_is_clear",
            "paint_corner",
            "corner_color_is",
        ]
        for mod in self.mods:
            for func in functions_to_override:
                setattr(mod, func, getattr(karel, func))

    def main(self) -> None:
        try:
            self.mods[0].main()
        except Exception as e:
            if isinstance(e, (KarelException, NameError, RuntimeError)):
                self.print_error_traceback(e)
            raise e

    def print_error_traceback(
        self, e: KarelException | NameError | RuntimeError
    ) -> None:
        """Handle runtime errors while executing student code."""
        display_frames: list[tuple[FrameType, int]] = []
        # Walk through all the frames in stack trace at time of failure
        for frame, lineno in tb.walk_tb(e.__traceback__):
            frame_info = inspect.getframeinfo(frame)
            # Get the name of the file corresponding to the current frame
            # Only display frames generated within the student's code
            if Path(frame_info.filename).name == f"{self.module_name}.py":
                display_frames.append((frame, lineno))

        display_frames_generator = (frame for frame in display_frames)
        trace = tb.format_list(tb.StackSummary.extract(display_frames_generator))
        clean_traceback = "".join(trace).strip()
        add_did_you_mean(e)
        print(
            f"Traceback (most recent call last):\n{clean_traceback}\n"
            f"{type(e).__name__}: {e}"
        )


class KarelApplication(tk.Frame):
    def __init__(
        self,
        karel: KarelProgram,
        code_file: Path,
        master: tk.Tk,
        window_width: int = 800,
        window_height: int = 600,
        canvas_width: int = 600,
        canvas_height: int = 400,
    ) -> None:
        # set window background to contrast white Karel canvas
        master.configure(background=LIGHT_GREY)

        # configure location of canvas to expand to fit window resizing
        master.rowconfigure(0, weight=1)
        master.columnconfigure(1, weight=1)

        # set master geometry
        master.geometry(f"{window_width}x{window_height}")

        super().__init__(master, background=LIGHT_GREY)

        self.karel = karel
        self.world = karel.world
        self.code_file = code_file
        self.load_student_code()
        master.title(self.student_code.module_name)
        if not self.student_code.mods:
            master.destroy()
            return
        self.icon = DEFAULT_ICON
        self.window_width = window_width
        self.window_height = window_height
        self.canvas_width = canvas_width
        self.canvas_height = canvas_height
        self.master = master
        self.set_dock_icon()
        self.grid(row=0, column=0)
        self.create_menubar()
        self.create_canvas()
        self.create_buttons()
        self.create_slider()
        self.create_status_label()

    def load_student_code(self) -> None:
        self.student_code = StudentCode(self.code_file)
        self.student_code.inject_namespace(self.karel)
        self.inject_decorator_namespace()

    def set_dock_icon(self) -> None:
        # make Karel dock icon image
        path = Path(__file__).absolute().parent / "icon.png"
        try:
            img = tk.Image("photo", file=path)
            # pylint: disable=protected-access
            self.master.tk.call("wm", "iconphoto", self.master._w, img)  # type: ignore[attr-defined] # noqa: E501
        except tk.TclError:
            print(f"Warning: invalid icon.png: {path}")

    def create_menubar(self) -> None:
        menubar = tk.Menu(self.master)
        file_menu = tk.Menu(menubar, tearoff=False)
        menubar.add_cascade(label="File", menu=file_menu)
        file_menu.add_command(
            label="Exit", underline=1, command=self.master.quit, accelerator="Cmd+W"
        )
        iconmenu = tk.Menu(menubar, tearoff=0)
        menubar.add_cascade(label="Select Icon", menu=iconmenu)
        iconmenu.add_command(label="Karel", command=lambda: self.set_icon("karel"))
        iconmenu.add_command(label="Simple", command=lambda: self.set_icon("simple"))

        self.bind_all("<Command-w>", lambda _: self.quit())
        self.master.config(menu=menubar)  # type: ignore[attr-defined]

    def create_slider(self) -> None:
        """
        This method creates a frame containing three widgets:
        two labels on either side of a scale slider to control
        Karel execution speed.
        """
        self.slider_frame = tk.Frame(self, bg=LIGHT_GREY)
        self.slider_frame.grid(row=3, column=0, padx=PAD_X, pady=PAD_Y, sticky="ew")

        self.fast_label = tk.Label(self.slider_frame, text="Fast", bg=LIGHT_GREY)
        self.fast_label.pack(side="right")

        self.slow_label = tk.Label(self.slider_frame, text="Slow", bg=LIGHT_GREY)
        self.slow_label.pack(side="left")

        self.speed = tk.DoubleVar()

        self.scale = tk.Scale(
            self.slider_frame,
            orient=tk.HORIZONTAL,
            variable=self.speed,
            showvalue=False,
        )
        self.scale.set(self.world.init_speed)
        self.scale.pack()

    def create_canvas(self) -> None:
        """This method creates the canvas on which Karel and the world are drawn."""
        self.canvas = KarelCanvas(
            self.canvas_width,
            self.canvas_height,
            self.master,
            world=self.world,
            karel=self.karel,
        )
        self.canvas.grid(column=1, row=0, sticky="NESW")
        self.canvas.bind("<Configure>", lambda t: self.canvas.redraw_all())

    def set_icon(self, icon: str) -> None:
        self.canvas.icon = icon
        self.canvas.redraw_karel()

    def create_buttons(self) -> None:
        """
        This method creates the three buttons that appear on the left
        side of the screen. These buttons control the start of Karel
        execution, resetting Karel's state, and loading new worlds.
        """
        self.program_control_button = tk.Button(
            self, highlightthickness=0, highlightbackground="white"
        )
        self.program_control_button["text"] = "Run Program"
        self.program_control_button["command"] = self.run_program
        self.program_control_button.grid(
            column=0, row=0, padx=PAD_X, pady=PAD_Y, sticky="ew"
        )

        self.load_world_button = tk.Button(
            self, highlightthickness=0, text="Load World", command=self.load_world
        )
        self.load_world_button.grid(
            column=0, row=2, padx=PAD_X, pady=PAD_Y, sticky="ew"
        )

    def create_status_label(self) -> None:
        """This function creates the status label at the bottom of the window."""
        self.status_label = tk.Label(
            self.master, text="Welcome to Karel!", bg=LIGHT_GREY
        )
        self.status_label.grid(row=1, column=0, columnspan=2)

    def karel_action_decorator(
        self, karel_fn: Callable[..., None]
    ) -> Callable[..., None]:
        def wrapper() -> None:
            # execute Karel function
            karel_fn()
            # redraw canvas with updated state of the world
            self.canvas.redraw_karel()
            # delay by specified amount
            sleep(1 - self.speed.get() / 100)

        return wrapper

    def beeper_action_decorator(
        self, karel_fn: Callable[..., None]
    ) -> Callable[..., None]:
        def wrapper() -> None:
            # execute Karel function
            karel_fn()
            # redraw canvas with updated state of the world
            self.canvas.redraw_beepers()
            self.canvas.redraw_karel()
            # delay by specified amount
            sleep(1 - self.speed.get() / 100)

        return wrapper

    def corner_action_decorator(
        self, karel_fn: Callable[..., None]
    ) -> Callable[..., None]:
        def wrapper(color: str) -> None:
            # execute Karel function
            karel_fn(color)
            # redraw canvas with updated state of the world
            self.canvas.redraw_corners()
            self.canvas.redraw_beepers()
            self.canvas.redraw_karel()
            # delay by specified amount
            sleep(1 - self.speed.get() / 100)

        return wrapper

    def inject_decorator_namespace(self) -> None:
        """
        This function is responsible for doing some Python hackery
        that associates the generic commands the student wrote in their
        file with specific commands relating to the Karel object that exists
        in the world.
        """
        for mod in self.student_code.mods:
            mod.turn_left = self.karel_action_decorator(self.karel.turn_left)
            mod.move = self.karel_action_decorator(self.karel.move)
            mod.pick_beeper = self.beeper_action_decorator(self.karel.pick_beeper)
            mod.put_beeper = self.beeper_action_decorator(self.karel.put_beeper)
            mod.paint_corner = self.corner_action_decorator(self.karel.paint_corner)

    def disable_buttons(self) -> None:
        self.program_control_button.configure(state="disabled")
        self.load_world_button.configure(state="disabled")

    def enable_buttons(self) -> None:
        self.program_control_button.configure(state="normal")
        self.load_world_button.configure(state="normal")

    def run_program(self) -> None:
        # Error checking for existence of main function completed in prior file

        # reimport code in case it changed
        self.load_student_code()
        try:
            self.status_label.configure(text="Running...", fg="brown")
            self.disable_buttons()
            self.student_code.main()
            self.status_label.configure(text="Finished running.", fg="green")

        except (KarelException, NameError):
            # Generate popup window to let the user know their program crashed
            self.status_label.configure(
                text="Program crashed, check console for details.", fg="red"
            )
            self.update()
            showwarning(
                "Karel Error", "Karel Crashed!\nCheck the terminal for more details."
            )

        finally:
            # Update program control button to force user
            # to reset world before running program again
            self.program_control_button["text"] = "Reset World"
            self.program_control_button["command"] = self.reset_world
            self.enable_buttons()

    def reset_world(self) -> None:
        self.karel.reset_state()
        self.world.reset_world()
        self.canvas.redraw_all()
        self.status_label.configure(text="Reset to initial state.", fg="black")
        # Once world has been reset, program control button resets to "run" mode
        self.program_control_button["text"] = "Run Program"
        self.program_control_button["command"] = self.run_program
        self.update()

    def load_world(self) -> None:
        default_worlds_path = Path(__file__).absolute().parent / "worlds"
        filename = askopenfilename(
            initialdir=default_worlds_path,
            title="Select Karel World",
            filetypes=[("Karel Worlds", "*.w")],
            parent=self.master,
        )
        # User hit cancel and did not select file, so leave world as-is
        if filename == "":
            return
        self.world.reload_world(filename=filename)
        self.karel.reset_state()
        self.canvas.redraw_all()
        # Reset speed slider
        self.scale.set(self.world.init_speed)
        self.status_label.configure(
            text=f"Loaded world from {Path(filename).name}.", fg="black"
        )

        # Make sure program control button is set to 'run' mode
        self.program_control_button["text"] = "Run Program"
        self.program_control_button["command"] = self.run_program



"""
This file defines the Karel ASCII program.

Original Author: Tyler Yep
License: MIT
Version: 1.0.0
Email: tyep@cs.stanford.edu
"""
from __future__ import annotations

from enum import Enum, unique
from typing import Any, Dict, Iterator, Tuple

from .karel_world import Direction, KarelWorld

CHAR_WIDTH = 5
HORIZONTAL, VERTICAL = "─", "│"
SPACING = 10
BEEPER_COORDS = Dict[Tuple[int, int], int]


class Tile:
    def __init__(self, value: str = "·") -> None:
        self.value = value
        self.walls: list[Direction] = []
        self.beepers = 0
        self.color = ""

    def __repr__(self) -> str:
        result = ""
        if self.value == "K" and self.beepers > 0:
            result = "<K>"
        elif self.beepers > 0:
            result = f"<{self.beepers}>"
        elif self.color:
            result = self.color[:3]
        else:
            result = self.value
        return result.center(CHAR_WIDTH)


class AsciiKarelWorld:
    def __init__(self, world: KarelWorld, karel_street: int, karel_avenue: int) -> None:
        num_sts, num_aves = world.num_streets, world.num_avenues
        # Initialize Tiles
        self.world_arr = [[Tile() for _ in range(num_aves)] for _ in range(num_sts)]

        # Add Karel
        self.world_arr[num_sts - karel_street][karel_avenue - 1].value = "K"
        for (avenue, street), count in world.beepers.items():
            self.world_arr[num_sts - street][avenue - 1].beepers = count

        # Add Walls
        for wall in world.walls:
            avenue, street, direction = wall.avenue, wall.street, wall.direction
            self.world_arr[num_sts - street][avenue - 1].walls.append(direction)

        # Add Colors
        for r in range(1, num_sts + 1):
            for c in range(1, num_aves + 1):
                if world.corner_color(c, r):
                    self.world_arr[num_sts - r][c - 1].color = world.corner_color(c, r)

        self.num_streets = num_sts
        self.num_avenues = num_aves

    def __repr__(self) -> str:
        avenue_widths = HORIZONTAL * ((CHAR_WIDTH + 1) * self.num_avenues + 1)
        result = f"┌{avenue_widths}┐\n"
        for r in range(self.num_streets):
            next_line = VERTICAL
            result += VERTICAL
            next_block_start = " "
            for c in range(self.num_avenues):
                tile = self.world_arr[r][c]
                line, next_block_start = self.get_next_line(r, c, next_block_start)
                next_line += line
                result += (
                    VERTICAL if self.tile_pair_has_wall(r, c, Direction.WEST) else " "
                ) + str(tile)

            result += f" {VERTICAL}\n"
            if r == self.num_streets - 1:
                result += f"└{avenue_widths}┘\n"
            else:
                result += f"{next_line} {VERTICAL}\n"
        return result

    def tile_has_wall(self, r: int, c: int, direction: Direction) -> bool:
        if 0 <= r < self.num_streets and 0 <= c < self.num_avenues:
            tile = self.world_arr[r][c]
            return direction in tile.walls
        return False

    def tile_pair_has_wall(self, r: int, c: int, direction: Direction) -> bool:
        """
        Checks whether the tile at r, c should have a wall by
        checking itself and its neighbors.
        """
        if direction == Direction.SOUTH:
            return self.tile_has_wall(r, c, Direction.SOUTH) or self.tile_has_wall(
                r + 1, c, Direction.NORTH
            )
        if direction == Direction.NORTH:
            return self.tile_has_wall(r, c, Direction.NORTH) or self.tile_has_wall(
                r - 1, c, Direction.SOUTH
            )
        if direction == Direction.WEST:
            return self.tile_has_wall(r, c, Direction.WEST) or self.tile_has_wall(
                r, c - 1, Direction.EAST
            )
        if direction == Direction.EAST:
            return self.tile_has_wall(r, c, Direction.EAST) or self.tile_has_wall(
                r, c + 1, Direction.WEST
            )
        raise ValueError("Direction is invalid.")

    def get_next_line(self, r: int, c: int, next_block_start: str) -> tuple[str, str]:
        """Given a tile, figures out the lower line of the ASCII art."""

        if self.tile_pair_has_wall(r, c, Direction.SOUTH):
            if (
                next_block_start == HORIZONTAL
                and self.tile_pair_has_wall(r, c, Direction.WEST)
                and self.tile_pair_has_wall(r + 1, c, Direction.WEST)
            ):
                next_block_start = "┼"
            elif (
                next_block_start == " "
                and self.tile_pair_has_wall(r, c, Direction.WEST)
                and self.tile_pair_has_wall(r + 1, c, Direction.WEST)
            ):
                next_block_start = "├"
            elif self.tile_pair_has_wall(r + 1, c, Direction.WEST):
                next_block_start = "┌"
            elif self.tile_pair_has_wall(r, c, Direction.WEST):
                next_block_start = "└"
            next_line = next_block_start + HORIZONTAL * CHAR_WIDTH
            next_block_start = HORIZONTAL
        else:
            if (
                next_block_start == HORIZONTAL
                and self.tile_pair_has_wall(r, c, Direction.WEST)
                and self.tile_pair_has_wall(r + 1, c, Direction.WEST)
            ):
                next_block_start = "┤"
            elif next_block_start == HORIZONTAL and self.tile_pair_has_wall(
                r + 1, c, Direction.WEST
            ):
                next_block_start = "┐"
            elif next_block_start == HORIZONTAL and self.tile_pair_has_wall(
                r, c, Direction.WEST
            ):
                next_block_start = "┘"
            elif self.tile_pair_has_wall(
                r, c, Direction.WEST
            ) and self.tile_pair_has_wall(r + 1, c, Direction.WEST):
                next_block_start = VERTICAL

            next_line = next_block_start + " " * CHAR_WIDTH
            next_block_start = " "
        return next_line, next_block_start


@unique
class Color(Enum):
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def compare_output(first: Any, second: Any) -> str:
    """Compares Karel Output and gets the results."""

    def create_two_column_string(col1: list[str], col2: list[str]) -> Iterator[str]:
        """col1 and col2 are Lists."""
        return map(lambda x: f"{x[0]}{' ' * SPACING}{x[1]}", zip(col1, col2))

    def symmetric_difference(
        a: BEEPER_COORDS, b: BEEPER_COORDS
    ) -> tuple[BEEPER_COORDS, BEEPER_COORDS]:
        extra_a, extra_b = {}, {}
        for k in a:
            if k not in b:
                extra_a[k] = a[k]
            elif a[k] - b[k] > 0:
                extra_a[k] = a[k] - b[k]
        for k in b:
            if k not in a:
                extra_b[k] = b[k]
            elif b[k] - a[k] > 0:
                extra_b[k] = b[k] - a[k]
        return extra_a, extra_b

    this, that = str(first).split("\n"), str(second).split("\n")
    world_width = len(this[0])

    header1, header2 = "Student Output:", "Expected Output:"
    text_spacing = " " * (world_width - len(header1) + SPACING + 1)
    two_columns = create_two_column_string(this, that)
    output = "\n".join(two_columns)
    fancy_arrows = f"{Color.RED.value}❯{Color.YELLOW.value}❯{Color.GREEN.value}❯ "

    result = (
        f"\n\n{fancy_arrows} "
        f"{Color.YELLOW.value}{first.world.world_file}{Color.END.value}"
        f"\n{header1}{text_spacing}{header2}\n{output}\n"
    )

    if first.avenue != second.avenue or first.street != second.street:
        result += (
            "Karel did not end up in the same location in both worlds:\n"
            f"Student: {(first.avenue, first.street)}\n"
            f"Expected: {(second.avenue, second.street)}\n\n"
        )
    if first.world.beepers != second.world.beepers:
        extra_a, extra_b = symmetric_difference(
            first.world.beepers, second.world.beepers
        )
        result += (
            "Beepers do not match: "
            "(Only beepers that appear in one world but not the other are listed)\n"
            f"Student: {extra_a}\n"
            f"Expected: {extra_b}\n\n"
        )
    return result



"""
This file defines the Style Checker for Karel programs.

Original Author: Tyler Yep
Credits: Brahm Capoor
License: MIT
Version: 1.0.0
Email: tyep@cs.stanford.edu
"""
import ast
import inspect
from pathlib import Path
from typing import Any, Callable

import stanfordkarel

from .karel_application import StudentCode


def style_test(func: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator for printing out the results of a style test."""

    def success_func(*args: Any, **kwargs: Any) -> bool:
        result = func(*args, **kwargs)
        if result:
            print("All good!")
        else:
            print("Error!")
        return result

    return success_func


class StyleChecker:
    """Style Checker for Karel."""

    def __init__(self, code_file: Path) -> None:
        self.student_code = StudentCode(code_file)
        self.modules = self.student_code.mods
        self.module_lines = str(self.student_code).split("\n")
        module_member_list = [inspect.getmembers(mod) for mod in self.modules]
        self.function_list = [
            var_name
            for var_name, value in
            # flatten module_member_list
            [m for module_members in module_member_list for m in module_members]
            if inspect.isfunction(value)
        ]

    @staticmethod
    def print_status_message(message: str) -> None:
        print(message.ljust(64), end="")

    def check_style(self) -> None:
        print(f"\n\nStyle Tests for {self.student_code.module_name}:")
        assert self.check_line_lengths()
        assert self.check_function_defs()
        assert self.assert_num_functions()
        assert self.check_recursion()
        assert self.check_naming()

    @style_test
    def check_recursion(self) -> bool:
        self.print_status_message("Checking for recursion...")
        return True

    @style_test
    def check_line_lengths(self, max_line_length: int = 88) -> bool:
        self.print_status_message("Checking line lengths...")
        long_idxs = [
            i for i, line in enumerate(self.module_lines) if len(line) > max_line_length
        ]
        if long_idxs:
            print(f"Lines {long_idxs} are longer than {max_line_length} characters.\n")
        return len(long_idxs) == 0

    @style_test
    def check_function_defs(self, min_name_length: int = 5) -> bool:
        self.print_status_message("Checking function definitions...")
        ok = True
        seen_already = set()
        for f in self.function_list:
            if f in seen_already:
                print(f"There are multiple functions defined that are called {f}")
                ok = False
            seen_already.add(f)
            if len(f) < min_name_length and f not in ("main", "move"):
                print(f"Function {f} has a pretty short name.")
                ok = False
        return ok

    @style_test
    def check_naming(self, min_name_length: int = 5) -> bool:
        """."""
        self.print_status_message("Checking function and variable names...")
        stanfordkarel_names = dir(stanfordkarel)
        root = ast.parse(str(self.student_code))
        names = sorted(
            {node.id for node in ast.walk(root) if isinstance(node, ast.Name)}
        )
        filtered_names = [name for name in names if name not in stanfordkarel_names]

        short_names = [
            f
            for f in filtered_names
            if len(f) < min_name_length and f not in ("main", "_")
        ]
        if short_names:
            print(f"Functions {short_names} have pretty short names.")
        return len(short_names) == 0

    @style_test
    def assert_num_functions(self, min_required: int = 10) -> bool:
        """
        Check that *at least* `num` functions
        are present in the module.
        """
        self.print_status_message("Checking number of functions...")
        num_fns = len(self.function_list)
        if num_fns < min_required:
            print(f"Expected at least {min_required} functions, only found {num_fns}.")
        return num_fns >= min_required
